from loginid import LoginID

# directweb application
BASE_URL = "http://localhost:8080/api/native"
CLIENT_ID ="l330ZWseeoJ3ooxHh0yz41s_3oYwYwVuTVt7J16djCgJ_56lYpMIXKPGpmYFQl3-trR34N0Wa3nHXpwZ8ga4wA=="
PRIVATE_KEY = """-----BEGIN PRIVATE KEY-----
MIGHAgEAMBMGByqGSM49AgEGCCqGSM49AwEHBG0wawIBAQQgwX3F5DgjQd9QlPmA
3wWtyoNltn5v+Gv02o8Zw4TF5ImhRANCAARH+9hX50YVg+TNVXSiCwN7thZBF5op
MHh9+toI5ZQPK0OdcuhbzVWuDpMRm9ivx3xAQBrRkCCDd5bG8Ba9lJo2
-----END PRIVATE KEY-----"""


# user info
USERNAME = "John.Doe.Quang"
USER_ID = "2ed409e8-e34c-4949-b074-4e0c2b52b0aa"

lAdmin = LoginID(CLIENT_ID, PRIVATE_KEY, base_url=BASE_URL)

# test helper functions
print("="*50)
print("Test `generateToken`")
print(lAdmin.generateServiceToken("codes.generate")) 

print("="*50)
print("test `decodeToken`")
token = "eyJhbGciOiJFUzI1NiIsImtpZCI6IjBiYTU5NTE0LTA5MDUtNDZjYS1hMDJhLWJkMWM2MDk4YjJhMCIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJsb2dpbmlkLmlvIiwic3ViIjoiMmVkNDA5ZTgtZTM0Yy00OTQ5LWIwNzQtNGUwYzJiNTJiMGFhIiwic2lkIjoiNTQ4NTdhMWQtZDI0Ni00Njc1LWE5MmItMzNjYmRkZjE5YmZjIiwibmlkIjoiNDAxYzRiZWViNzkyNTg1MiIsImF1ZCI6ImwzMzBaV3NlZW9KM29veEhoMHl6NDFzXzNvWXdZd1Z1VFZ0N0oxNmRqQ2dKXzU2bFlwTUlYS1BHcG1ZRlFsMy10clIzNE4wV2EzbkhYcHdaOGdhNHdBPT0iLCJhY3Rpb24iOiJsb2dpbiIsImlhdCI6MTYyMzI2Nzk3NiwidWRhdGEiOiJKb2huLkRvZS5RdWFuZyJ9.zNwZ3j8YhfTHkzPDJkfR24Q0nYJ8roqEzbEmeACCZ0XmwJdlzLANfxcmD4tthB57pwrHvkXpsuWoNpdl9QEr4w"
print(lAdmin.verifyToken(token))

# Code APIs
# common settings
codeType = 'short'
codePurpose = 'temporary_authentication'

# waitCode function will stall the test flow. 
# enable this variable to test waitCode
TEST_WAIT_CODE = False

print("="*50)
print("test `generateCode`")
err, code_response = lAdmin.generateCode(USER_ID, codeType, codePurpose, False)

if err is not None:
    print(err, code_response)
else:
    print(code_response)

print("="*50)
print("test `authorizeCode`")
err, response = lAdmin.authorizeCode(USER_ID, code_response['code'], codeType, codePurpose)

if err is not None:
    print(err, response)
else:
    print(response)

print("="*50)
print("test `invalidateAllCodes`")
# generate some code to invalidate
err, code_response = lAdmin.generateCode(USER_ID, "short", "temporary_authentication", False)

if err is not None:
    print('failed to generate code')
    print(err, code_response)
else:
    print('code generated', code_response)
    print('invalidate code')
    print(lAdmin.invalidateAllCodes(USER_ID, codeType, codePurpose))

print("="*50)
if TEST_WAIT_CODE:

    print("test `waitCode`")
    # generate some code to invalidate
    err, code_response = lAdmin.generateCode(USER_ID, "short", "temporary_authentication", False)

    if err is not None:
        print('failed to generate code')
        print(err, code_response)
    else:
        print('code generated', code_response)
        print('waiting for code to be authorized')
        err, response = lAdmin.waitCode(USERNAME, code_response['code'], codeType)
        print(response)
else:
    print("TEST_WAIT_CODE is False, skipping waitCode test")

# Transaction Flow
sample_payload = "A sample payload"
print("="*50)
print("test `generateTxAuthToken`")
print(lAdmin.generateTxAuthToken(sample_payload, USERNAME))

print("="*50)
print("test `createTxId`")
err, response = lAdmin.createTxId(sample_payload, USERNAME)
if err is not None:
    print(err, response)
else:
    print('tx_id', response)


from os import error
import requests, json, jwt

import random
from datetime import datetime
import base64, hashlib

from typing import Tuple, Union, Dict

from requests.api import head
class LoginID:
    def __init__(self, client_id: str, private_key: str, base_url: str) -> None:
        """ This server SDK leverages either a web or mobile application 
        and requires an API credential to be assigned to that integration.

        Args:
            client_id (str): The client ID
            private_key (str): API private key
            base_url (str): LoginID serivce address

        Example:: 

            from loginid import LoginID
            lid = LoginID(CLIENT_ID, PRIVATE_KEY)
        
           
        """
        self._private_key = private_key
        self._client_id = client_id
        self._base_url = base_url

        # allowed code types
        self._code_types = {'short', 'long', 'phrase'}

    def get_clientId(self) -> str:
        """Extract the client id 

        Returns:
            str: The client id
        """
        return self.client_id

    def _getUtcEpoch(self):
        return int((datetime.utcnow()-datetime(1970,1,1)).total_seconds())

    def _getRandomString(self, length: int=16) -> str:
        """
        Generate a random string of given length with alphanumeric characters

        Args:
            length (int): length of the output string

        Returns:
            str: a random string
        """
        possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
        ret = [random.choice(possible) for _ in range(length)]
        return "".join(ret)

    def _getPublicKey(self, kid: str) -> Tuple[Union[Exception, None], str]:
        """Extract the client's public key with `kid`

        Args:
            kid (str): the kid included in jwt's header

        Returns:
            tuple: a 2-element tuple containing

                - **error** (*int*): HTTP error code, `None` if no errors
                - **response** (str): The public key if `error` is `None`, else the error message
        """
        try:
            params = {"kid":kid}
            r = requests.get(f'{self._base_url}/certs', params=params)
            if (r.status_code != 200): return r.status_code, r.json()
            
            return None, r.text 
        except Exception as e:
            return e, str(e)
        
    def verifyToken(self, token: str, username: Union[None, str]=None) -> bool:
        """Verify a JWT token returned upon user authorization

        Args:
            token (str): JWT token
            username (Union[None, str], optional): if given, checks for if `username` matches the `udata` in JWT.
                 Defaults to None.

        Returns:
            bool: `True` if the token is valid, `False` otherwise (including errors)
        """

        try:
            headers = jwt.get_unverified_header(token)
            algo, kid = headers['alg'], headers['kid']
            assert algo=="ES256", f"{algo} is not an allowed algorithm."

            # extracting public key from LogInID
            err, public_key = self._getPublicKey(kid)
            if err is not None:
                return False

            payload = jwt.decode(token, public_key, 
                algorithms=algo,
                audience=self._client_id
            )

            if username is not None:
                return username == payload['udata']
            return True
        except:
            return False

    def generateServiceToken(self, scope: str, algo: str="ES256", username: str=None, user_id:str=None, nonce: str=None) -> str:
        """
        Generate a service token

        Args:
            scope (str): the scope of the service
            algo (str, optional): Encryption algorithm, defaults to `"ES256"`
            username (str, optional): the username to-be granted by the token
            user_id (str, optional): the user_id to be granted by the token. If `username` is given, 
                this is ignored
            nonce (str, optional): nonce for the token, randomly generated if not given

        Returns:
            str: the JWT service token
        """
        payloads = {
            "client_id": self._client_id,
            "type": scope,
            "nonce": nonce or self._getRandomString(16),
            "iat": self._getUtcEpoch()
        }

        if username is not None:
            payloads['username'] = username
        elif user_id is not None:
            payloads['user_id'] = user_id

        jwt_headers = {"alg": algo, "typ": "JWT"}

        return jwt.encode(
            payloads, self._private_key, algorithm=algo,
            headers=jwt_headers
        )

    def _request(self, method: str, url: str, payloads: dict, headers:dict=None) -> Tuple[Union[None, int], Dict]:
        methods = {'post': requests.post, 'get': requests.get}
        r_method = methods.get(method, None)
        if r_method is None:
            return -1, {'message': f'request method {method} not allowed'}
        
        r = r_method(url, data=json.dumps(payloads), headers=headers)

        if (r.status_code != 200):
            return r.status_code, r.json()
        
        return None, r.json()
        
    # Tx Flow 
    def generateTxAuthToken(self, tx_payload: str, username: str=None, nonce: str=None) -> str:
        """Generate an Authorization Token for Transaction Flow

        Args:
            tx_payload (str): The transaction payload
            username (str, optional): The username
            nonce (str, optional): optional nonce for the token, auto-generated if not given

        Returns:
            str: The JWT authorization token
        """

        # encryption algorithm, only supports ES256
        algo = "ES256"

        # hash and encode tx_payload
        payload_hash = hashlib.sha256(tx_payload.encode()).digest()
        payload_hash = (base64.urlsafe_b64encode(payload_hash)
                              .decode().strip('=')
                       )

        payloads = {
            'type': 'tx.create',
            'nonce': nonce or self._getRandomString(16),
            'payload_hash': payload_hash,
            'iat': self._getUtcEpoch()
        }

        # TODO: remove username when native-service is updated
        if username is not None: payloads['username'] = username

        return jwt.encode(
            payloads, self._private_key,
            algorithm=algo,
            headers={'alg': algo, 'typ': 'JWT'}
        )

    def createTx(self, tx_payload: str, username: str=None, nonce: str=None) -> Tuple[Union[Exception, int, None], str]:
        """Create a Transaction ID

        Args:
            tx_payload (str): The transaction payload
            username (str, optional): The username that initiates the transaction
            nonce (str, optional): The optional nonce, randomly generated if not provided


        Returns:
            tuple: a 2-element tuple containing

                - **error** (*int*): The HTTP error, `None` if no errors            
                - **tx_id** (*str*): The transaction id if no error, the empty string otherwise
        """
        try:
            url = f"{self._base_url}/tx"

            headers = {
                "Content-type": "application/json",
                "Authorization": "Bearer " + self.generateTxAuthToken(tx_payload)
            }
            payloads = dict(
                client_id=self._client_id,
                username=username,
                tx_type='text',
                tx_payload=tx_payload,
                nonce=nonce or self._getRandomString()  
            )

            err, response = self._request('post', url, payloads, headers=headers)

            if (err is not None):
                return err, ""

            return None, response['id']
        except Exception as e:
            return e, str(e) 

    def verifyTransaction(self, tx_token: str, tx_payload: str) -> Tuple[Union[Exception, None], bool]: 
        """Verify the jwt token returned upon completion of a transaction

        Args:
            tx_token (str): The JWT token 
            tx_payload (str): the original transaction payload 

        Returns:
            tuple: A 2-element tuple containing

                **error** (*Exception*): Error code if errors occur, `None` otherwise
                **valid** (*bool*): `True` if the token is valid, `False` if not
        """
        try:
            headers = jwt.get_unverified_header(tx_token)
            algo, kid = headers['alg'], headers['kid']
            assert algo == "ES256", f"{algo} is not an allowed algorithm"

            # get public key
            err, public_key = self._getPublicKey(kid)
            if err is not None:
                return err, public_key

            payload = jwt.decode(tx_token, public_key, algorithms=algo,
                                 audience=self._client_id)

            to_hash = tx_payload + payload.get("nonce","") + payload.get("server_nonce","")
            hash = hashlib.sha256(to_hash.encode()).digest()
            hash = base64.urlsafe_b64encode(hash).decode().strip('=')

            return None, payload.get('tx_hash') == hash
        except Exception as e:
            return e, str(e)
            
    # Code
    def isValidCodeType(self, codeType: str)->bool:
        """Check if a code type is valid

        Args:
            codeType (str): code type

        Returns:
            bool: `True` if codeType is value, `False` otherwise
        """
        return codeType in self._code_types

    def waitCode(self, username: str, code: str, codeType: str)->Tuple[Union[int, Exception, None], dict]:
        """ Wait for a given code

        Args:
            username (str): The username
            code (str): The code associate to the username
            codeType (str): Type of the code

        Returns:
            tuple: a 2-element tuple containing

                - **error** (*int*): The HTTP error, `None` if no errors
                - **response** (*dict*): The response body from code wait 
        """
        try:
            assert self.isValidCodeType(codeType)

            url = f'{self._base_url}/authenticate/code/wait'

            headers = {
                "Content-type": "application/json",
                "Authorization": "Bearer " + self.generateServiceToken("auth.temporary")
            }

            payloads = {
                "client_id": self._client_id,
                "username": username,
                "authentication_code": {
                    "code" : code,
                    "type": codeType
                }
            }

            r = requests.post(url, data=json.dumps(payloads), headers=headers)

            if (r.status_code != 200):
                return r.status_code, r.json()

            json_response = r.json()
            token = json_response.get('jwt')
            if (token is None) or (not self.verifyToken(token)):
                raise Exception("Cannot authorize code")

            return None, r.json()
        except Exception as e:
            return e, {"message": str(e)}



from os import error
import requests, json, jwt

import random
from datetime import datetime
import base64, hashlib

from typing import Tuple, Union, Dict

from .loginid import LoginID

class LoginIdManagement(LoginID):
    def __init__(self, client_id: str, private_key: str, base_url: Union[str, None]=None) -> None:
        """This server SDK can be used with a management application and requires an API credential to be assigned to that integration. 
        All calls made from this SDK are intended to be backend-to-backend calls, as the operations are sensitive.

        Args:
            client_id (str): Client ID as per the value provided on the dashboard
            private_key (str): Private key of the API credential assigned to the application on the dashboard.
            base_url (Union[str, None], optional): Base URL where requests should be made (if specifying which environment is being used). Defaults to the LoginID production URL if not given (`None`)

        Example::

            from loginid import LoginIdManagememnt
            lManagement = LoginIdManagement(CLIENT_ID, PRIVATE_KEY)
        """
        base_url = base_url or "https://usw1.loginid.io/"
        super().__init__(client_id, private_key, base_url=base_url)

    # User Management    
    def getUserId(self, username: str) -> Tuple[Union[Exception,int], Dict]:
        """Get the user ID from username

        Args:
            username (str): The username

        Returns:
            tuple: a 2-element tuple with
            
                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): the user id if no errors, else the JSON response
        """
        try:
            url = self._base_url + "/manage/users/retrieve"

            headers = {
                "Content-type": "application/json",
                "x-client-id": self._client_id,
                "Authorization": "Bearer " + self.generateServiceToken("users.retrieve")
            }

            payloads = {
                "username": username
            }

            r = requests.post(url, data=json.dumps(payloads), headers=headers)

            if (r.status_code != 200):
                return r.status_code, r.json()
            
            return None, r.json()['id']
        except Exception as e:
            return e, {'error': str(e)}

    def deleteByUsername(self, username: str) -> Tuple[Union[int, Exception], bool]:
        """Delete a User by the username

        Args:
            username (str): the username to be deleted

        Returns:
            tuple: a 2-element tuple with
            
                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): success message if user deleted successfully, else the JSON response
        """
        try:
            url = self._base_url + "/manage/users/delete"
            headers = {
                "Content-type": "application/json",
                "x-client-id": self._client_id,
                "Authorization": "Bearer " + self.generateServiceToken("users.delete")
            } 

            payloads = {
                "username": username
            }

            r = requests.post(url=url, data=json.dumps(payloads), headers=headers)

            if r.status_code != 204:
                print(r.status_code)
                return r.status_code, r.json()
            return None, {"message": f"User {username} deleted"}
        except Exception as e:
            return e, {'error':str(e)}

    def deleteByUserId(self, user_id: str) -> Tuple[Union[None, Exception], Dict]:
        """Delete a user by user Id

        Args:
            user_id (str): the user id to be deleted

        Returns:
            tuple: a 2-element tuple with
            
                - **error** (*Union[int, Exception]*): HTTP error code if any, Exception for unknown error, `None` if no errors
                - **response** (*Union[str, Dict]*): success message if user deleted successfully, else the JSON response
        """
        try:
            url = f"{self._base_url}/manage/users/{user_id}"

            headers = {
                "Content-type":"application/x-www-form-urlencoded",
                "Authorization": "Bearer " + self.generateServiceToken("users.delete"),
                "x-client-id": self._client_id
            }

            r = requests.delete(url, headers=headers)
            
            if r.status_code != 204:
                return r.status_code, r.json()
            return None, {'message': f'user with id {user_id} deleted'}
        except Exception as e:
            return e, {'error': str(e)}

    def activateUserById(self, user_id: str) -> Tuple[Union[None, Exception], Dict]:
        """Activate a previously deactivated user

        Args:
            user_id (str): the user id to be activated

        Returns:
            tuple: a 2-element tuple with
            
                - **error** (*Union[int, Exception]*): HTTP error code if any, Exception for unknown error, `None` if no errors
                - **response** (*Union[str, Dict]*): user profiles if no error, else the JSON response
        """
        try:
            url = f"{self._base_url}/manage/users/{user_id}/activate"

            headers = {
                "Content-type": "application/x-www-form-urlencoded",
                "Authorization": "Bearer " + self.generateServiceToken("users.activate"),
                "x-client-id": self._client_id
            }

            r = requests.put(url, headers=headers)

            if r.status_code != 200:
                return r.status_code, r.json()

            return None, r.json()
        except Exception as e:
            return e, {'error': str(e)}

    def deactivateUserById(self, user_id: str) -> Tuple[Union[None, Exception], Dict]:
        """Deactivate a currently active user

        Args:
            user_id (str): the user id to be deactivated

        Returns:
            tuple: a 2-element tuple with
            
                - **error** (*Union[int, Exception]*): HTTP error code if any, Exception for unknown error, `None` if no errors
                - **response** (*Union[str, Dict]*): user profiles if no error, else the JSON response
        """
        try:
            url = f"{self._base_url}/manage/users/{user_id}/deactivate"

            headers = {
                "Content-type": "application/x-www-form-urlencoded",
                "Authorization": "Bearer " + self.generateServiceToken("users.deactivate"),
                "x-client-id": self._client_id
            }

            r = requests.put(url, headers=headers)

            if r.status_code != 200:
                return r.status_code, r.json()

            return None, r.json()         
        except Exception as e:
            return e, {'error': str(e)}

    # Code APIs
    def generateCode(self, user_id: str, codeType: str, 
                     codePurpose: str, isAuthorized: bool) -> Tuple[Union[None,int], Dict]:
        """Generate a code

        Args:
            user_id (str): The user_id for the code
            codeType (str): The code type
            codePurpose (str): The purpose of the code
            isAuthorized (bool): Whether the code authorizes the user or not


        Returns:
            tuple: a 2-element tuple with 
            
                - **error** (*int*): HTTP error code, `None` if no errors
                - **response** (*dict*): the response body from code generation request
        """
        url = f"{self._base_url}/codes/{codeType}/generate"

        headers = {
            "Content-type": "application/json",
            "Authorization": "Bearer " + self.generateServiceToken("codes.generate")
        }
        payloads = dict(
            client_id=self._client_id,
            user_id=user_id,
            purpose=codePurpose,
            authorize=isAuthorized
        )

        r = requests.post(url, data=json.dumps(payloads), headers=headers)

        if (r.status_code != 200):
            return r.status_code, r.json()

        return None, r.json()

    def authorizeCode(self, user_id: str, code: str, 
                      codeType: str, codePurpose: str) -> Tuple[Union[None,int], Dict]:
        """Authorize a given code

        Args:
            user_id (str): The user_id associate with the code
            code (str): The code that needs authorization
            codeType (str): The type of the code
            codePurpose (str): The purpose of the code

        Returns:
            tuple: a 2-element tuple containing

                - **error** (*int*): The HTTP error, `None` if no errors
                - **response** (*dict*): The response body from code authorization
        """
        try:
            assert self.isValidCodeType(codeType), f"{codeType} is not a valid code type"

            url = f"{self._base_url}/codes/{codeType}/authorize"

            headers = {
                "Content-type": 'application/json',
                'Authorization': 'Bearer ' + self.generateServiceToken('codes.authorize')
            }

            payloads = dict(
                client_id=self._client_id,
                user_id=user_id,
                purpose=codePurpose,
                code=code
            )

            r = requests.post(url, data=json.dumps(payloads), headers=headers)

            # TODO: throw a proper error here
            if (r.status_code != 200):
                return r.status_code, r.json()

            return None, r.json()
        except Exception as e:
            return e, {'message': str(e)}
    
    def invalidateAllCodes(self, user_id: str, codeType: str, codePurpose: str) -> int:
        """Invalidate the given code

        Args:
            user_id (str): the user ID
            codeType (str): code type
            codePurpose (str): the purpose of the code 

        Returns:
            int: `0` if the code is successfully invalidated, `-1` for unknown error, otherwise HTTP response status code  
        """
        try:
            assert self.isValidCodeType(codeType)

            url = f'{self._base_url}/codes/{codeType}/invalidate-all'

            headers = {
                "Content-type": "application/json",
                "Authorization": "Bearer " + self.generateServiceToken("codes.invalidate")
            }

            payloads = dict(
                client_id=self._client_id,
                user_id=user_id,
                purpose=codePurpose,
            )
            r = requests.post(url, data=json.dumps(payloads), headers=headers)
            
            if r.status_code != 200: return r.status_code

            return 0
        except:
            return -1

    # Credentials
    def getCredentials(self, user_id: str) -> Tuple[Union[int, Exception], Dict]:
        """Get an exhaustive list of credentials for a given user

        Args:
            user_id (str): User ID of the end user who you would like to get the list of credentials

        Returns:
            tuple: A 2-element tuple consisting of

                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): User's credentials if no error, else the JSON response
        """
        try:
            url = f"{self._base_url}/credentials"

            headers = {
                "Content-type": "application/x-www-form-urlencoded",
                "Authorization" : "Bearer " + self.generateServiceToken("credentials.list"),
                "x-api-key": self._client_id
            }

            params = {
                "user_id": user_id
            }

            r = requests.get(url, params=params, headers=headers)

            if r.status_code != 200:
                return r.status_code, str(r)

            return None, r.json()
        except Exception as e:
            return e, str(e)
    
    def renameCredential(self, user_id: str, cred_id: str, updated_name: str) -> Tuple[Union[int, Exception], Dict]:
        """Rename a credential of a user

        Args:
            user_id (str): The id of the user
            cred_id (str): The id of the credential to be renamed
            updated_name (str): The new name

        Returns:
            tuple: A 2-element tuple consisting of

                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): The renamed credential's detail if no error, else the JSON response
        """
        try:
            url = self._base_url + "/credentials/rename"

            headers = {
                "Content-type": "application/json",
                "Authorization": "Bearer " + self.generateServiceToken("credentials.rename"),
                # "x-api-key": self._client_id
            }

            payloads = {
                "client_id": self._client_id,
                "user_id": user_id,
                "credential": {
                    "uuid": cred_id,
                    "name": updated_name
                }
            }

            r = requests.post(url, data=json.dumps(payloads), headers=headers)
            
            if r.status_code!=200:
                return r.status_code, r.json()
            
            return None, r.json()
        except Exception as e:
            return e, str(e)

    def revokeCredential(self, user_id: str, cred_id: str) -> Tuple[Union[int, Exception], Dict]:
        """Revoke an existing credential from a user

        Args:
            user_id (str): The user id to extract the credential
            cred_id (str): The credential id to be revoked

        Returns:
            tuple: A 2-element tuple consisting of

                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): The revoked credential's detail if no error, else the JSON response
        """
        try:
            url = self._base_url + "/credentials/revoke"
            headers = {
                "Content-type": "application/json",
                "Authorization": "Bearer " + self.generateServiceToken("credentials.revoke")
            }
            payloads = {
                "client_id": self._client_id,
                "user_id": user_id,
                "credential":{
                    "uuid": cred_id
                }
            }
            r = requests.post(url, data=json.dumps(payloads), headers=headers)

            if r.status_code != 200:
                return r.status_code, r.json()

            return None, r.json()
        except Exception as e:
            return e, str(e)

    def addUserWithoutCredentials(self, username: str) -> Tuple[Union[int, Exception], Dict]:
        """Add a new user without credentials. The new user can create new credentials with recovery flow

        Args:
            username (str): The username of the new user

        Returns:
            tuple: A 2-element tuple consisting of

                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): The new user's profile if no errors, else the JSON response

        """
        try:
            url = f'{self._base_url}/manage/users'

            headers = {
                "Content-type": "application/json",
                "Authorization": "Bearer " + self.generateServiceToken("users.create"),
                "x-client-id": self._client_id
            }

            payloads = {
                "username": username
            }

            r = requests.post(url, data=json.dumps(payloads), headers=headers)

            if (r.status_code !=200):
                return r.status_code, r.json()

            return None, r.json()
        except Exception as e:
            return e, {'message': str(e)}
            
    def initAddCredentialWithoutCode(self, user_id: str) -> Tuple[Union[int, Exception], Dict]:
        """Add a credential without pre-generated authorization code
        Args:
            user_id (str): ID of the user to add new credential to
        Returns:
            tuple: A 2-element tuple consisting of
                - **error** (*Union[int, Exception]*): HTTP error code, `None` if no errors
                - **response** (*Union[str, Dict]*): The attestation payload for the credentials if no errors, else the JSON response
        """
        url = f'{self._base_url}/credentials/fido2/init/force'

        headers = {
            "Content-type": "application/json",
            "Authorization": f"Bearer {self.generateServiceToken('credentials.force_add')}"
        }

        payloads = {
            "client_id": self._client_id,
            "user_id": user_id
        }
        
        r = requests.post(url, data=json.dumps(payloads), headers=headers)

        if (r.status_code != 200):
            return r.status_code, r.json()

        return None, r.json()